import datetime
from datetime import timedelta
import json

import requests
from cacheops import cached
from django.apps import apps
from django.conf import settings
from django.utils.translation import gettext_lazy as _
from rest_framework_simplejwt.exceptions import TokenBackendError, TokenError
from rest_framework_simplejwt.tokens import UntypedToken
from rest_framework_simplejwt.utils import aware_utcnow


def aiit_get_token_by_login(username, password):
    """
    通过用户名密码获取token
    :param username: 用户名
    :param password: 密码
    :return: 返回 token 的值
    """
    url = '{}/authority/manage/user/login'.format(settings.SERVER_URL)
    payload = json.dumps({
        "password": password,
        "username": username
    })
    headers = {
        'Content-Type': 'application/json'
    }
    response = requests.request("POST", url, headers=headers, data=payload)

    res_json = response.json()
    token = ''
    if res_json:
        try:
            token = res_json.get('result', {}).get('token', '')
        except AttributeError:
            pass
    return token


@cached(timeout=60*10)
def token_validate(jwt_token):
    """
    验证token是否有效
    :param jwt_token: 'Bearer ...'
    :return: 返回 True 或者 False
    """
    url = "{}/authority/token/validate".format(settings.SERVER_URL)
    payload = json.dumps({"token": jwt_token})
    headers = {
        'Authorization': jwt_token,
        'Content-Type': 'application/json'
    }
    response = requests.request("POST", url, headers=headers, data=payload)
    res_json = response.json()
    is_valid = res_json.get('result', {}).get('isValid', False)
    return is_valid


@cached(timeout=60*30)
def token_decode(jwt_token):
    """
    解析 token 的信息，通过 token 获取到 username 之后，如果此用户没有在本系统，系统会自动为本用户创建账号
    :param jwt_token: 'Bearer ...'
    :return: {'username': '', 'user_id': 1, 'exp': 1644917927}
    """
    url = "{}/authority/token/decode".format(settings.SERVER_URL)
    payload = json.dumps({"token": jwt_token})
    headers = {
        'Authorization': jwt_token,
        'Content-Type': 'application/json'
    }
    response = requests.request("POST", url, headers=headers, data=payload)
    res_json = response.json()
    payload = res_json.get('result', {})
    username = payload.get('username')
    if username:
        app_name, model_name = settings.AUTH_USER_MODEL.split('.')
        user_model = apps.get_app_config(app_name).get_model(model_name)
        aiit_user, created = user_model.objects.get_or_create(username=username)
        payload['user_id'] = aiit_user.pk
    payload['exp'] = 3600 + int(datetime.datetime.now().timestamp())
    return payload


class AiitToken(UntypedToken):
    lifetime = timedelta(seconds=0)

    def __init__(self, token=None, verify=True):
        self.token = token
        self.current_time = aware_utcnow()

        # Set up token
        if token is not None:
            # # Decode token
            try:
                self.payload = token_decode('Bearer {}'.format(token.decode()))
            except TokenBackendError:
                raise TokenError(_('Token is invalid or expired'))

            if verify:
                self.verify()

    def verify(self):
        is_valid = token_validate('Bearer {}'.format(self.token.decode()))
        if not is_valid:
            raise TokenError(_('token 无效'))



from rest_framework.pagination import PageNumberPagination
from rest_framework.response import Response
from rest_framework.utils.urls import replace_query_param, remove_query_param


class NormalResultsSetPagination(PageNumberPagination):
    page_size = 10
    page_size_query_param = 'page_size'
    max_page_size = 100

    def get_next_url(self):
        if not self.page.has_next():
            return None
        url = self.request.get_full_path()
        page_number = self.page.next_page_number()
        return replace_query_param(url, self.page_query_param, page_number)

    def get_previous_url(self):
        if not self.page.has_previous():
            return None
        url = self.request.get_full_path()
        page_number = self.page.previous_page_number()
        if page_number == 1:
            return remove_query_param(url, self.page_query_param)
        return replace_query_param(url, self.page_query_param, page_number)

    def get_paginated_response(self, data):
        if not self.page.has_next():
            next_page = None
        else:
            next_page = self.page.next_page_number()
        return Response({
            'next_url': self.get_next_url(),
            'previous_url': self.get_previous_url(),
            'next_link': self.get_next_link(),
            'previous_link': self.get_previous_link(),
            'next_page': next_page,
            'current_length': len(data),
            'total_count': self.page.paginator.count,
            'page_num': self.page.number,
            'num_pages': self.page.paginator.num_pages,
            'page_size': self.page.paginator.per_page,
            'data': data,
            'message': 'ok',
            'code': 200
        })


class LargeNormalResultsSetPagination(NormalResultsSetPagination):
    page_size = 20


class AiitPagination(NormalResultsSetPagination):
    page_size = 10
    page_size_query_param = 'page_size'
    max_page_size = 100

    def get_paginated_response(self, data):
        return Response(
            data={
                'data': data,
                'page': {
                    'currentPage': self.page.number,
                    'totalPage': self.page.paginator.num_pages,
                    'pageSize': self.page.paginator.per_page,
                    'totalSize': self.page.paginator.count,
                    'hasNext': self.page.has_next(),
                    'hasPrevious': self.page.has_previous()
                },
                'message': 'ok',
                'code': 200
            })


class LargeAiitPagination(AiitPagination):
    page_size = 20



import json

from django.contrib.admin.models import LogEntry, ADDITION, CHANGE, DELETION
from django.contrib.contenttypes.models import ContentType


def get_content_type_for_model(obj):
    return ContentType.objects.get_for_model(obj, for_concrete_model=False)


def field_value_to_json(f_value):
    if str(type(f_value)) in [
        "<class 'django.db.models.fields.files.FieldFile'>",
        "<class 'django.db.models.fields.files.ImageFile'>",
        "<class 'django.db.models.fields.files.ImageFieldFile'>",
        "<class 'django.core.files.uploadedfile.InMemoryUploadedFile'>"
    ]:
        value = f_value.name
    elif str(type(f_value)) in [
        "<class 'uuid.UUID'>",
        "<class 'datetime.datetime'>",
        "<class 'datetime.date'>",
        "<class 'bool'>"
    ]:
        value = str(f_value)
    elif 'models' in str(type(f_value)):
        value = f_value.pk
    elif str(type(f_value)) in [
        "<class 'decimal.Decimal'>"
    ]:
        value = str(float(f_value))
    else:
        value = str(f_value)
    return value


def create_addition_log(user_id, oj):
    lg = LogEntry.objects.log_action(
        user_id=user_id,
        content_type_id=get_content_type_for_model(oj).pk,
        object_id=oj.pk,
        object_repr=str(oj),
        action_flag=ADDITION,
    )
    # 因为 LogEntry 不会保存具体数据，需要另外将操作涉及的数据存入该条记录中
    addition_message_dict = {
        'addition': {
            'fields': list(),
            'values': list()
        }
    }
    for i in oj._meta.fields:
        fields = i.attname
        value = getattr(oj, fields)
        addition_message_dict['addition']['fields'].append(fields)
        addition_message_dict['addition']['values'].append(field_value_to_json(value))
    addition_message = json.dumps(addition_message_dict, ensure_ascii=False)
    lg.change_message = addition_message
    lg.save()
    return True


def create_change_log(user_id, oj, origin_data, new_data):
    lg = LogEntry.objects.log_action(
        user_id=user_id,
        content_type_id=get_content_type_for_model(oj).pk,
        object_id=oj.pk,
        object_repr=str(oj),
        action_flag=CHANGE,
    )
    change_message_dict = {
        'changed': {
            'fields': list(),
            'origin_values': list(),
            'new_values': list(),
        }
    }
    for fields, value in new_data.items():
        origin_value = origin_data.get(fields)
        if value != origin_value and fields in origin_data:
            change_message_dict['changed']['fields'].append(fields)
            change_message_dict['changed']['origin_values'].append(field_value_to_json(origin_value))
            change_message_dict['changed']['new_values'].append(field_value_to_json(value))
    change_message = json.dumps(change_message_dict, ensure_ascii=False)
    lg.change_message = change_message
    lg.save()
    return True


def create_delete_log(user_id, oj):
    lg = LogEntry.objects.log_action(
        user_id=user_id,
        content_type_id=get_content_type_for_model(oj).pk,
        object_id=oj.pk,
        object_repr=str(oj),
        action_flag=DELETION,
    )
    # 因为 LogEntry 不会保存具体数据，需要另外将操作涉及的数据存入该条记录中
    delete_message_dict = {
        'delete': {
            'fields': list(),
            'values': list()
        }
    }
    for i in oj._meta.fields:
        fields = i.attname
        value = getattr(oj, fields)
        delete_message_dict['delete']['fields'].append(fields)
        delete_message_dict['delete']['values'].append(field_value_to_json(value))
    delete_message = json.dumps(delete_message_dict, ensure_ascii=False)
    lg.change_message = delete_message
    lg.save()
    return True



import copy
from decimal import Decimal

from django.db.models import ManyToManyField, DateField, TimeField, ForeignKey, DateTimeField
from django_filters.rest_framework import FilterSet
from rest_framework.generics import ListAPIView, RetrieveAPIView, RetrieveUpdateAPIView, \
    CreateAPIView, ListCreateAPIView, RetrieveUpdateDestroyAPIView

from .response import APIResponse
from .log import create_addition_log, create_change_log, create_delete_log


class AiitFilter(FilterSet):

    @classmethod
    def filter_for_field(cls, field, field_name, lookup_expr='exact'):
        filter_class = super().filter_for_field(field, field_name, lookup_expr)
        if lookup_expr == 'exact':
            filter_class.extra['help_text'] = '{0}等于'.format(field.verbose_name)
        elif lookup_expr == 'contains':
            filter_class.extra['help_text'] = '{0}包含'.format(field.verbose_name)
        elif lookup_expr == 'gte':
            filter_class.extra['help_text'] = '{0}大于等于'.format(field.verbose_name)
        elif lookup_expr == 'gt':
            filter_class.extra['help_text'] = '{0}大于'.format(field.verbose_name)
        elif lookup_expr == 'lt':
            filter_class.extra['help_text'] = '{0}小于'.format(field.verbose_name)
        elif lookup_expr == 'lte':
            filter_class.extra['help_text'] = '{0}小于等于'.format(field.verbose_name)
        return filter_class


def convert_obj_to_dict(obj, fields=None, exclude=None):
    """
    将 Model 对象 obj 转化为字典格式返回
    :param obj: 通过 model.objects.filter().first() 、model.objects.get() 或者 for i in QuerySet 获取的对象
    :param fields: 要选择性呈现的字段
    :param exclude: 排除呈现的字段
    :return: 返回字典 {'id': 1, 'field1': 'field1_data'}
    """
    data = {}
    for f in getattr(obj, '_meta').concrete_fields + getattr(obj, '_meta').many_to_many:
        value = f.value_from_object(obj)

        if fields and f.name not in fields:
            continue

        if exclude and f.name in exclude:
            continue

        if isinstance(f, ManyToManyField):
            value = [i.id for i in value] if obj.pk else None

        elif isinstance(f, DateTimeField):
            value = value.strftime('%Y-%m-%d %H:%M:%S') if value else None

        elif isinstance(f, DateField):
            value = value.strftime('%Y-%m-%d') if value else None

        elif isinstance(f, TimeField):
            value = value.strftime('%H:%M:%S') if value else None

        elif isinstance(f, Decimal):
            value = float(f)

        # ForeignKey 特殊处理
        if isinstance(f, ForeignKey):
            data[f.column] = value
            data[f.name] = value
        else:
            data[f.name] = value

    # 获取 property 里面的数据
    for p in getattr(getattr(obj, '_meta'), '_property_names'):
        value = getattr(obj, p)
        if isinstance(value, (str, int, Decimal)):
            data[p] = value
    return data


def obj_list(self):
    queryset = self.filter_queryset(self.get_queryset())

    page = self.paginate_queryset(queryset)
    if page is not None:
        serializer = self.get_serializer(page, many=True)
        return self.get_paginated_response(serializer.data)

    serializer = self.get_serializer(queryset, many=True)
    return APIResponse(serializer.data)


def obj_create(self, request):
    new_data = copy.deepcopy(request.data)
    request_user = request.user
    user_id = None
    if request_user:
        user_id = request_user.id
        if user_id:
            new_data['created_by'] = user_id
            new_data['updated_by'] = user_id
    serializer = self.get_serializer(data=new_data)
    serializer.is_valid(raise_exception=True)
    oj = serializer.save()
    create_addition_log(user_id=user_id, oj=oj)
    headers = self.get_success_headers(serializer.data)
    return APIResponse(serializer.data, headers=headers)


def obj_retrieve(self):
    instance = self.get_object()
    serializer = self.get_serializer(instance)
    return APIResponse(data=serializer.data)


def obj_update(self, request, *args, **kwargs):
    partial = kwargs.pop('partial', False)
    instance = self.get_object()
    origin_data = convert_obj_to_dict(instance)
    new_data = copy.deepcopy(request.data)
    request_user = request.user
    user_id = None
    if request_user:
        user_id = request_user.id
        if user_id:
            user_id = user_id
            new_data['updated_by'] = request_user.id
    serializer = self.get_serializer(instance, data=new_data, partial=partial)
    serializer.is_valid(raise_exception=True)
    oj = serializer.save()
    create_change_log(user_id, oj, origin_data, new_data)

    if getattr(instance, '_prefetched_objects_cache', None):
        instance._prefetched_objects_cache = {}

    return APIResponse(data=serializer.data)


def obj_delete(self, request):
    instance = self.get_object()
    request_user = request.user
    user_id = None
    if request_user:
        user_id = request_user.id
    create_delete_log(user_id, oj=instance)
    self.perform_destroy(instance)
    return APIResponse(status=204)


class AiitListAPIView(ListAPIView):

    def list(self, request, *args, **kwargs):
        return obj_list(self)


class AiitCreateAPIView(CreateAPIView):

    def create(self, request, *args, **kwargs):
        return obj_create(self, request)


class AiitListCreateAPIView(ListCreateAPIView):

    def list(self, request, *args, **kwargs):
        return obj_list(self)

    def create(self, request, *args, **kwargs):
        return obj_create(self, request)


class AiitRetrieveAPIView(RetrieveAPIView):
    filter_backends = []

    def retrieve(self, request, *args, **kwargs):
        return obj_retrieve(self)


class AiitRetrieveUpdateAPIView(RetrieveUpdateAPIView):
    filter_backends = []

    def retrieve(self, request, *args, **kwargs):
        return obj_retrieve(self)

    def update(self, request, *args, **kwargs):
        return obj_update(self, request, *args, **kwargs)

    def post(self, request, *args, **kwargs):
        return self.partial_update(request, *args, **kwargs)


class AiitRetrieveUpdateDestroyAPIView(RetrieveUpdateDestroyAPIView):
    filter_backends = []

    def retrieve(self, request, *args, **kwargs):
        return obj_retrieve(self)

    def update(self, request, *args, **kwargs):
        return obj_update(self, request, *args, **kwargs)

    def destroy(self, request, *args, **kwargs):
        return obj_delete(self, request)

    def post(self, request, *args, **kwargs):
        return self.partial_update(request, *args, **kwargs)



from rest_framework.response import Response


class APIResponse(Response):

    def __init__(self, data=None, message='ok', code=200, status=200, headers=None, exception=False, **kwargs):
        # res_data 的初始状态：状态码与状态信息
        res_data = {
            'data': data,
            'message': message,
            'code': code
        }
        res_data.update(kwargs)
        # 重写父类的Response的__init__方法
        super().__init__(data=res_data, status=status, headers=headers, exception=exception)



import json
from pathlib import Path

import requests
from django.conf import settings


def get_algo_url(algo_name):
    """
    通过算法名称获取算法调用的API
    :param algo_name: 算法名称
    :return: 返回算法调用的API
    """
    search_url = '{}/api/algo/procedure/?ordering=-id&name={}'.format(settings.SERVER_URL, algo_name)
    response = requests.request("GET", search_url)
    response_data = response.json()
    data = response_data.get('data', [])
    if data:
        algo_data = data[0]
        if algo_data:
            algo_url = algo_data.get('api_url')
            if not algo_url.startswith('http'):
                algo_url = '{}{}'.format(settings.SERVER_URL, algo_url)
            return algo_url
    return ''


def exec_algo(algo_name, **params):
    """
    执行算法
    :param algo_name: 算法的名称
    :param params: 算法的输入参数
    :return: 以字典的形式返回算法的执行结果
    """
    algo_url = get_algo_url(algo_name)
    params['async'] = False
    if algo_url:
        payload = json.dumps(params)
        headers = {'Content-Type': 'application/json'}
        response = requests.request("POST", algo_url, headers=headers, data=payload)
        if response.status_code == 200:
            response_data = response.json()
            return response_data
    else:
        raise ValueError
    return {}


def upload_file(file_path):
    url = "{}/api/algo/file_upload/".format(settings.SERVER_URL)
    file_path_obj = Path(file_path)
    file_type_dict = {
        '.txt': 'text/plain',
        '.csv': 'text/csv',
        '.json': 'application/json',
        '.pdf': 'application/pdf',
        '.zip': 'application/zip',
        '.png': 'image/png',
        '.jpg': 'image/jpeg',
        '.md': 'text/plain',
    }
    file_type = file_type_dict.get(file_path_obj.suffix, 'text/plain')
    files = [
        ('file', (file_path_obj.name, open(str(file_path_obj.absolute()), 'rb'), file_type))
    ]
    response = requests.request("POST", url, files=files)
    res = response.json()
    return res.get('file', '')



import datetime
import os
from pathlib import Path

from .response import APIResponse
from django.core.files.storage import FileSystemStorage
from rest_framework.views import APIView


class AiitStorage(FileSystemStorage):

    def url(self, name):
        res = super().url(name=name)
        if res:
            res = '{}?tool={}'.format(res, os.environ.get('DJANGO_SETTINGS_MODULE').split('.')[0])
        return res


def save_request_files(original_files):
    """
    存储文件并返回文件存储路径
    """
    files_new = dict()
    if not Path('media', 'tmp').exists():
        Path('media', 'tmp').mkdir(exist_ok=True)
    for param_name, file_obj in original_files.items():
        file_obj_name = '{}_{}{}'.format(
            Path(file_obj.name).stem,
            datetime.datetime.now().strftime('%Y%m%d%H%M%S'),
            Path(file_obj.name).suffix
        )
        file_path = str(Path('media', 'tmp', file_obj_name))
        with open(file_path, 'wb+') as destination:
            for chunk in file_obj.chunks():
                destination.write(chunk)
        files_new[param_name] = '/{}?tool={}'.format(file_path, os.environ.get('DJANGO_SETTINGS_MODULE').split('.')[0])
    return files_new


class FileUploadView(APIView):
    """
    post:

        文件上传接口。在调用算法之前将文件先上传到服务器中。
        可以上传单个或者多个文件，参数名称没有要求。

    """

    def post(self, request):
        files_new = save_request_files(request.FILES)
        return APIResponse(data=files_new)



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name='aiit_sdk',
    version='0.0.7',
    author='Guoqun Jin',
    author_email='guoqun.jin@hotmail.com',
    description='A python sdk for AIIT OS',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='http://192.168.140.202/unstructured-etl-group/aiit-sdk.git',
    license='GNU General Public License v3.0',
    project_urls={
        "Bug Tracker": "http://192.168.140.202/unstructured-etl-group/aiit-sdk/-/issues",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    package_dir={"": "src"},
    packages=setuptools.find_packages(where="src"),
    install_requires=[
        "requests>=2.25.1",
        "django-cacheops>=5.1",
        "django-filter>=21.1",
        "djangorestframework>=3.12.2",
        "djangorestframework-simplejwt>=4.6.0",
    ],
    python_requires=">=3.7"
)
